/*
 *
 * Copyright  :: 1995-2006 (C)opyright Gordano Ltd
 *               All rights reserved.
 *               No copies of this source may be made without express written
 *               premission of Gordano Ltd, 18 Kenn Road, Clevedon, BS21 6EL
 *
 * $Header: d:\source\data\ntmail17_01\optpack\efmarchive/efm.c,v 1.2 2006/05/25 10:53:01 simont Exp $
 *
 * $Log: efm.c,v $
 * Revision 1.2  2006/05/25  10:53:01  simont
 * Added new API
 *
 * Revision 1.1  2006/01/06  09:57:54  simont
 * Initial revision
 *
 *
 */

/*
 * Example EFM Archiver DLL for Windows.
 *
 * This EFM Archive DLL implements an external message archive for GMS. Archived messages are stored in
 * the C:\EFM directory. Messages are stored in single files in standard RFC822 format including the
 * trailing ".". 
 * 
 * There are 2 options for archiving depending on the requirements:
 *
 * 1. Archive data that would normally go into the GMS archive (MesLog).
 * 2. Archive data that would go into an end user folder.
 *  
 * These are slightly different when aliases and SMTP rejections are taken into account.
 *
 * This DLL implements both.
 *
 * The file structure is:
 *
 *		C:\EFM\<year>\<month>\<day>\<domain>\<user>\<folder>\<uid>.mbx
 *
 * Additional files are kept to optimise the nextuid and uidvalidity calculations.
 *
 * Known limitations:
 *
 * 1. This is a read only archive.
 * 2. A real archive would probably index data for searching.
 *
 */

#if defined( WINNT ) || defined( WIN95 )
#pragma warning ( disable : 4001 4115 4201 4209 4214 4514 )
#include <windows.h>
#pragma warning ( disable : 4001 4115 4201 4209 4214 4514 4100 4127 4152 4706)
#else
#include "windows.h"
#endif

#include <stdio.h>

#ifndef MAIN
#define MAIN
#endif

#include "efm.h"

/*
 * GMS error codes
 */
#define S_SUCCESS                   0

#define E_OUT_MEMORY                26
#define E_WRITE_FILE_FAILURE        33
#define E_READ_FILE_FAILURE         34
#define E_NO_FILE                   56
#define E_BAD_VALUE                 112
#define E_CREATE_DIR				114
#define E_NO_FOLDER                 125
#define W_NOT_HANDLED				-219
#define E_INTERNAL_ERROR            231
#define E_NOT_IMPLEMENTED           241

/*
 * Filenames and locations
 */
#define EFMBaseDir				"C:\\EFMArchive"
#define EFMNextUIDFilename		"NextUID.txt"
#define EFMUIDValidityFilename  "UIDValidity.txt"

#define EFM_MAILBOX_EXT			"mbx"
#define EFM_ATTRIBUTES_EXT		"idx"

/*
 * Timeout on opening files
 */
#define EFM_LOCK_TIMEOUT 30


/*
 * Miscellaneous
 */
#define EFM_MAX_PARAMETER   1024
#define EFM_MAX_LINE        10240

/*
 * Private Types
 */
typedef struct {
	char user[EFM_MAX_EMAIL_ADDRESS];
	char domain[EFM_MAX_EMAIL_ADDRESS];
	SYSTEMTIME now;
	int serviceType;
} EFM_SESSION;

typedef struct {
	char user[EFM_MAX_EMAIL_ADDRESS];
	char domain[EFM_MAX_EMAIL_ADDRESS];
	char foldername[EFM_MAX_FILENAME];
	char indexfilename[EFM_MAX_FILENAME];
	int serviceType;
	int count;
} EFM_FOLDER;

/*
 * Entry and Exit Trace
 */
static LPFN_EFM_UPDATE_CALLBACK_FN efmUpdateCallback = NULL;

void *LogHandle = NULL;
void (*DLLLogMes) (void *Log, int Level, int id, const char *fmt, ...) = NULL;

BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
	if (reason == DLL_PROCESS_ATTACH)
	{
		return TRUE;
	}
	else if (reason == DLL_PROCESS_DETACH)
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

char *CreateUDIL(char *result, size_t size)
{
	static LONG g = 0;
	SYSTEMTIME now;

	GetSystemTime(&now);

	InterlockedIncrement(&g);

	sprintf(result, "%02d%02d%02d%03d%05d", now.wHour, now.wMinute, now.wSecond, now.wMilliseconds, g % 100000);

	return result;
}

static HANDLE RetryCreateFile(LPCTSTR filename, DWORD dwDesiredAccess, DWORD dwShareMode,
			      DWORD dwCreationDistribution, DWORD dwFlagsAndAttributes, int iRetryCount)
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	int attempt = 0;

	if (iRetryCount <= 0)
		iRetryCount = 1;
	iRetryCount *= 10;

	while (attempt < iRetryCount && hFile == INVALID_HANDLE_VALUE)
	{
		hFile = CreateFile(filename, dwDesiredAccess, dwShareMode, NULL,
				   dwCreationDistribution, dwFlagsAndAttributes, NULL);
		attempt++;
		if (iRetryCount > 10 && hFile == INVALID_HANDLE_VALUE)
		{
			DWORD error = GetLastError();

			if (error == ERROR_PATH_NOT_FOUND)
			{
				char Directory[EFM_MAX_FILENAME];
				char *End;

				strcpy(Directory, filename);
				End = strrchr(Directory, '\\');
				if (End)
				{
					*End = '\0';

					CreateDirectory(Directory, NULL);
				}
			}

			if (attempt < iRetryCount)
				Sleep(100);
		}
	}

	return (hFile);
}

static int RecurseCreateDirectory(const char *path)
{
	char Buffer[EFM_MAX_FILENAME];
	char *End;
	BOOL OK = TRUE;

	strcpy(Buffer, path);

	End = Buffer + strlen(Buffer) - 1;
	if (*End == '\\')
		*End = '\0';

	End = Buffer;
	if (*End == '\\')
		End++;
	End = strchr(End, '\\');
	if (End)
		if (*(End + 1) == '\\' || *(End - 1) == ':')
			End = strchr(End + 2, '\\');

	if (End)
		*End = '\0';

	while (End)
	{
		OK = CreateDirectory(Buffer, NULL);

		*End = '\\';
		End++;

		End = strchr(End, '\\');

		if (End)
			*End = '\0';
	}

	OK = CreateDirectory(Buffer, NULL);

	return (OK);
}

static int EFMGetUIDValidity(int id, SYSTEMTIME * now, char *user, char *domain, char *foldername, BOOL reset,
			     unsigned int *uidvalidity)
{
	BOOL ok;
	int status = S_SUCCESS;
	HANDLE hFile;
	char filename[EFM_MAX_FILENAME];
	char buffer[EFM_MAX_PARAMETER];
	DWORD len = sizeof(buffer);

	*uidvalidity = 0;

	sprintf(filename, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s\\%s", EFMBaseDir, now->wYear, now->wMonth, now->wDay, domain, user,
		foldername, EFMUIDValidityFilename);

	if (reset)
	{
		DeleteFile(filename);
	}

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		ok = ReadFile(hFile, buffer, sizeof(buffer), &len, FALSE);

		if (ok)
		{
			if (len != 0)
			{
				*(buffer + len) = '\0';

				*uidvalidity = atoi(buffer);
			}
			else
			{
				SYSTEMTIME st;

				GetSystemTime(&st);

				len = sprintf(buffer, "%02u%1u%02u%05u",
					      st.wMonth, st.wYear % 10, st.wDay, (st.wHour * 60 + st.wMinute) * 60 + st.wSecond);

				*uidvalidity = atoi(buffer);

				SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
				SetEndOfFile(hFile);

				ok = WriteFile(hFile, buffer, len, &len, FALSE);

				if (!ok)
				{
					status = E_WRITE_FILE_FAILURE;
				}
			}
		}
		else
		{
			status = E_READ_FILE_FAILURE;
		}

		CloseHandle(hFile);

		if (status != S_SUCCESS)
		{
			DeleteFile(filename);
		}
	}
	else
	{
		status = E_NO_FILE;
	}

	return status;
}

static int EFMGetNextUID(int id, SYSTEMTIME * now, char *user, char *domain, char *foldername, BOOL allocate, unsigned int *uid)
{
	BOOL ok;
	int status = S_SUCCESS;
	HANDLE hFile;
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	char filename[EFM_MAX_FILENAME];
	char buffer[EFM_MAX_PARAMETER];
	DWORD len = sizeof(buffer);
	unsigned int fileuid = 0;
	unsigned int maxuid = 0;
	BOOL save = allocate;
	unsigned int uidvalidity;

	*uid = 0;

	sprintf(filename, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s\\%s", EFMBaseDir, now->wYear, now->wMonth, now->wDay, domain, user,
		foldername, EFMNextUIDFilename);

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		ok = ReadFile(hFile, buffer, sizeof(buffer), &len, FALSE);

		if (ok)
		{
			if (len != 0)
			{
				*(buffer + len) = '\0';

				*uid = atoi(buffer);
			}
			else
			{
				// Rebuild the max_uid from the files
				sprintf(filename, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s\\*.%s", EFMBaseDir, now->wYear, now->wMonth,
					now->wDay, domain, user, foldername, EFM_MAILBOX_EXT);

				hFindFile = FindFirstFile(filename, &finddata);

				if (hFindFile != INVALID_HANDLE_VALUE)
				{
					ok = TRUE;

					while (ok)
					{
						fileuid = atoi(finddata.cFileName);

						if (fileuid > maxuid)
						{
							maxuid = fileuid;
						}

						ok = FindNextFile(hFindFile, &finddata);
					}

					FindClose(hFindFile);

					*uid = maxuid + 1;
				}
				else
				{
					*uid = 1;
				}

				// Need to reset UIDValidity
				status = EFMGetUIDValidity(id, now, user, domain, foldername, TRUE, &uidvalidity);

				save = TRUE;
			}
		}
		else
		{
			status = E_READ_FILE_FAILURE;
		}

		if (status == S_SUCCESS)
		{
			maxuid = *uid;

			if (allocate)
			{
				maxuid++;
			}

			if (save)
			{
				len = sprintf(buffer, "%d", maxuid);

				SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
				SetEndOfFile(hFile);

				ok = WriteFile(hFile, buffer, len, &len, FALSE);

				if (!ok)
				{
					status = E_WRITE_FILE_FAILURE;
				}
			}
		}

		CloseHandle(hFile);

		if (status != S_SUCCESS)
		{
			DeleteFile(filename);
		}
	}
	else
	{
		status = E_NO_FILE;
	}

	return status;
}

static void EFMGetMailboxFilenameFromUID(SYSTEMTIME * now, char *user, char *domain, char *foldername, unsigned int uid,
					 char *filename, int size)
{
	sprintf(filename, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s\\%d.%s", EFMBaseDir, now->wYear, now->wMonth, now->wDay, domain, user,
		foldername, uid, EFM_MAILBOX_EXT);
}

static void EFMGetAttributesFilenameFromUID(SYSTEMTIME * now, char *user, char *domain, char *foldername, unsigned int uid,
					    char *filename, int size)
{
	sprintf(filename, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s\\%d.%s", EFMBaseDir, now->wYear, now->wMonth, now->wDay, domain, user,
		foldername, uid, EFM_ATTRIBUTES_EXT);
}

static int EFMCleanupUIDFiles(SYSTEMTIME * now, char *user, char *domain, char *foldername, unsigned int uid)
{
	int status = S_SUCCESS;
	char filename[EFM_MAX_FILENAME];

	EFMGetMailboxFilenameFromUID(now, user, domain, foldername, uid, filename, sizeof(filename));
	DeleteFile(filename);

	EFMGetAttributesFilenameFromUID(now, user, domain, foldername, uid, filename, sizeof(filename));
	DeleteFile(filename);

	return status;
}

static int EFMWriteAttributes(SYSTEMTIME * now, char *user, char *domain, char *foldername, int uid, char *udil, char *flags)
{
	BOOL ok;
	int status = S_SUCCESS;
	char filename[EFM_MAX_FILENAME];
	HANDLE hFile;
	char buffer[EFM_MAX_LINE];
	DWORD len;

	EFMGetAttributesFilenameFromUID(now, user, domain, foldername, uid, filename, sizeof(filename));

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		len = sprintf(buffer, "%s %s", udil, flags);

		SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
		SetEndOfFile(hFile);

		ok = WriteFile(hFile, buffer, len, &len, FALSE);

		if (!ok)
		{
			status = E_WRITE_FILE_FAILURE;
		}

		CloseHandle(hFile);
	}
	else
	{
		status = E_NO_FILE;
	}

	return status;
}

static void EFMFreeFolderList(EFM_FOLDER_LIST ** list)
{
	EFM_FOLDER_LIST *current;
	EFM_FOLDER_LIST *next;

	current = *list;

	while (current != NULL)
	{
		next = current->next;

		if (current->foldername != NULL)
		{
			free(current->foldername);
		}

		free(current);

		current = next;
	}

	*list = NULL;
}

int EFMConfigure(int id, void *logMes, void *logHandle)
{
	int status = E_BAD_VALUE;

	if (logMes != NULL)
	{
		DLLLogMes = logMes;
		LogHandle = logHandle;

		status = S_SUCCESS;
	}

	return (status);
}

int EFMClose(int id)
{
	int status = S_SUCCESS;

	// No action

	return status;
}

int EFMCreateSession(int id, char *user, char *domain, int serviceType, EFM_SESSION_HANDLE * sessionHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	*sessionHandle = EFM_INVALID_SESSION_HANDLE;

	session = malloc(sizeof(EFM_SESSION));

	if (session != NULL)
	{
		strcpy(session->user, user);
		strcpy(session->domain, domain);
		GetLocalTime(&session->now);
		session->serviceType = serviceType;

		*sessionHandle = (EFM_SESSION_HANDLE) session;
	}
	else
	{
		status = E_OUT_MEMORY;
	}

	return status;
}

int EFMCloseSession(int id, EFM_SESSION_HANDLE * sessionHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (*sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) * sessionHandle;

		free((void *)*sessionHandle);
		*sessionHandle = EFM_INVALID_SESSION_HANDLE;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMKeepAlive(int id, EFM_SESSION_HANDLE sessionHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMExpungedMessageAccessAllowed(int id, EFM_SESSION_HANDLE sessionHandle, BOOL *allowed)
{
	int status = S_SUCCESS;

	*allowed = TRUE;

	return status;
}

/*
 * Messages
 */
int EFMGetMessage(int id, EFM_FOLDER_HANDLE folderHandle, unsigned int uid, char *filename, size_t size)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseMessage(int id, EFM_FOLDER_HANDLE folderHandle, unsigned int uid, char *filename)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCopyMessages(int id, EFM_FOLDER_HANDLE folderHandle, char *foldername, EFM_MESSAGE_LIST * messageList)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMSetMessageFlags(int id, EFM_FOLDER_HANDLE folderHandle, unsigned int uid, char *flags)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

/*
 * Folders
 */
int EFMIsExternalFolder(int id, char *user, char *domain, int serviceType, BOOL *external)
{
	int status = S_SUCCESS;

	*external = FALSE;

	// Only archive SMTP data
	if (serviceType == EFM_SERVICE_TYPE_SMTP)
	{
		// Everything is external. We'll worry about what to actually save later,
		*external = TRUE;
	}

	return status;
}

int EFMGetFolderUIDValidity(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, unsigned int *uidvalidity)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMGetUIDValidity(id, &session->now, session->user, session->domain, foldername, FALSE, uidvalidity);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderNextUID(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, unsigned int *uid)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMGetNextUID(id, &session->now, session->user, session->domain, foldername, FALSE, uid);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}


int EFMExistFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, BOOL *exists)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	WIN32_FIND_DATA finddata;
	HANDLE hFindFile;
	char filename[EFM_MAX_FILENAME];

	*exists = FALSE;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		sprintf(filename, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s", EFMBaseDir, session->now.wYear, session->now.wMonth,
			session->now.wDay, session->domain, session->user, foldername);

		hFindFile = FindFirstFile(filename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
			{
				*exists = TRUE;
			}

			FindClose(hFindFile);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCreateFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	BOOL exists = FALSE;
	char directory[EFM_MAX_FILENAME];
	unsigned int uidvalidity;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMExistFolder(id, sessionHandle, foldername, &exists);

		if (status == S_SUCCESS && !exists)
		{
			sprintf(directory, "%s\\%02d\\%02d\\%02d\\%s\\%s\\%s", EFMBaseDir, session->now.wYear, session->now.wMonth,
				session->now.wDay, session->domain, session->user, foldername);

			if (RecurseCreateDirectory(directory))
			{
				// Set the UID Validity up
				status =
				    EFMGetUIDValidity(id, &session->now, session->user, session->domain, foldername, FALSE,
						      &uidvalidity);
			}
			else
			{
				status = E_CREATE_DIR;
			}
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMDeleteFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMFreeFolderList(list);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderModified(int id, EFM_SESSION_HANDLE sessionHandle, char *folderName, char *modifiedStr, int size)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderSize(int id, EFM_SESSION_HANDLE sessionHandle, char *folderName, DWORD * size)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMFolderCountMessagesOfType(int id,
				 EFM_SESSION_HANDLE sessionHandle,
				 char *foldername, char *includeFlags, char *excludeFlags, int *messageCount)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMRenameFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, char *newfoldername, BOOL renameSubFolders)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;


	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMOpenFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, int flags, EFM_FOLDER_HANDLE * folderHandle)
{
	int status = S_SUCCESS;

	if (flags == EFM_FOLDER_OPEN_FULL)
	{
		*folderHandle = EFM_INVALID_FOLDER_HANDLE;

		if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
		{
			// Not required for SMTP
			status = E_NOT_IMPLEMENTED;
		}
		else
		{
			status = E_INTERNAL_ERROR;
		}
	}
	else if (flags == EFM_FOLDER_OPEN_INFORMATION)
	{
		if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
		{
			// Not required for SMTP
			status = E_NOT_IMPLEMENTED;
		}
		else
		{
			status = E_INTERNAL_ERROR;
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderIndex(int id, EFM_FOLDER_HANDLE folderHandle, EFM_MESSAGE_LIST ** messageList)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	*messageList = NULL;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseFolderIndex(int id, EFM_FOLDER_HANDLE folderHandle, EFM_MESSAGE_LIST ** messageList)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;
	EFM_MESSAGE_LIST *currentMessageList = NULL;
	EFM_MESSAGE_LIST *nextMessageList = NULL;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		currentMessageList = *messageList;

		while (currentMessageList != NULL)
		{
			nextMessageList = currentMessageList->next;

			free(currentMessageList);

			currentMessageList = nextMessageList;
		}

		*messageList = NULL;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseFolder(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_HANDLE * folderHandle, int flags)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;
	}

	if (*folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCommitFolder(int id, EFM_FOLDER_HANDLE folderHandle)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMPurgeFolder(int id, EFM_FOLDER_HANDLE folderHandle)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMStoreFolder(int id,
		   EFM_SESSION_HANDLE sessionHandle,
		   char *foldername, char *filename, char *flags, unsigned int *uid, char *udil, int size)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_SESSION *session;
	unsigned int newuid = 0;
	char newudil[EFM_MAX_PARAMETER];
	char mailboxFilename[EFM_MAX_FILENAME];
	char *ptr;
	BOOL copy = FALSE;
	BOOL archive = FALSE;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Decide if we wish to handle this message or not.
		if (*session->user != '\0' && *session->domain != '\0')
		{
			// User folder
			copy = TRUE;
		}
		else
		{
			ptr = strchr(foldername, '\\');

			if (ptr != NULL)
			{
				*ptr = '\0';

				if (stricmp(foldername, "MesLog") == 0)
				{
					// Archive folder
					archive = TRUE;
				}
			}
		}

		if (copy || archive)
		{
			status = EFMCreateFolder(id, sessionHandle, foldername);

			if (status == S_SUCCESS)
			{
				status =
				    EFMGetNextUID(id, &session->now, session->user, session->domain, foldername, TRUE, &newuid);
			}

			if (status == S_SUCCESS)
			{
				EFMGetMailboxFilenameFromUID(&session->now, session->user, session->domain, foldername, newuid,
							     mailboxFilename, sizeof(mailboxFilename));

				ok = CopyFile(filename, mailboxFilename, TRUE);

				if (!ok)
				{
					status = E_WRITE_FILE_FAILURE;
				}
			}

			if (status == S_SUCCESS)
			{
				CreateUDIL(newudil, sizeof(newudil));

				status =
				    EFMWriteAttributes(&session->now, session->user, session->domain, foldername, newuid, newudil,
						       flags);
			}

			if (status == S_SUCCESS)
			{
				if (uid != NULL)
				{
					*uid = newuid;
				}

				if (udil != NULL)
				{
					strcpy(udil, newudil);
				}
			}
			else
			{
				EFMCleanupUIDFiles(&session->now, session->user, session->domain, foldername, newuid);
			}
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	if (status == S_SUCCESS)
	{
		if (copy)
		{
			// We have taken a copy. Allow GMS to deliver as normal.
			status = W_NOT_HANDLED;
		}
	}

	return status;
}

/*
 * Folder Locking
 */
int EFMClientFolderLock(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, EFM_LOCK_HANDLE *lockHandle)
{
	int status = S_SUCCESS;

	// No locking
	*lockHandle = EFM_INVALID_LOCK_HANDLE;

	return status;
}

int EFMServiceFolderLock(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, EFM_LOCK_HANDLE *lockHandle)
{
	int status = S_SUCCESS;

	// No locking
	*lockHandle = EFM_INVALID_LOCK_HANDLE;

	return status;
}

int EFMCloseFolderLock(int id, EFM_LOCK_HANDLE *lockHandle)
{
	int status = S_SUCCESS;

	// No locking
	*lockHandle = EFM_INVALID_LOCK_HANDLE;

	return status;
}

int EFMGMSClientFolderLockRequired(int id, BOOL *required)
{
	int status = S_SUCCESS;

	// We do no locking so we should require GMS to do the locking
	*required = TRUE;

	return status;
}

int EFMGMSServiceFolderLockRequired(int id, BOOL *required)
{
	int status = S_SUCCESS;

	// We do no locking so we should require GMS to do the locking
	*required = TRUE;

	return status;
}

int EFMDoesClientFolderLockExist(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, BOOL *exists)
{
	int status = S_SUCCESS;

	// No locking
	*exists = FALSE;

	return status;
}

int EFMDoesServiceFolderLockExist(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, BOOL *exists)
{
	int status = S_SUCCESS;

	// No locking
	*exists = FALSE;

	return status;
}

/*
 * Folder Subscription
 */
int EFMSubscribeFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMUnsubscribeFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetSubscribedFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseSubscribedFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// Not required for SMTP
		status = E_NOT_IMPLEMENTED;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

/*
 * Update callback
 */
int EFMRegisterUpdateCallback(int id, LPFN_EFM_UPDATE_CALLBACK_FN updateCallback)
{
	int status = S_SUCCESS;

	// Not required for SMTP
	status = E_NOT_IMPLEMENTED;

	return status;
}
