/*
 *
 * Copyright  :: 1995-2006 (C)opyright Gordano Ltd
 *               All rights reserved.
 *               No copies of this source may be made without express written
 *               premission of Gordano Ltd, 18 Kenn Road, Clevedon, BS21 6EL
 *
 * $Header: d:\source\data\ntmail17_01\optpack\efmfull/efm.c,v 1.2 2006/05/25 10:53:21 simont Exp $
 *
 * $Log: efm.c,v $
 * Revision 1.2  2006/05/25  10:53:21  simont
 * Added new API
 *
 * Revision 1.1  2006/01/18  08:18:26  simont
 * Initial revision
 *
 *
 */

/*
 * Example EFM DLL for Windows.
 *
 * This EFM DLL implements an external message store for GMS. Messages are stored in
 * the C:\EFM directory. Messages are stored in single files in standard RFC822 format including the
 * trailing ".". The file structure is:
 *
 *			C:\EFM\<domain>\<user>\<folder>\<uid>.mbx
 *
 * Additional files are kept to optimise the nextuid and uidvalidity calculations.
 *
 * Known limitations:
 *
 * 1. Message statuses are not cached and the commit API is not implemented.
 * 2. Should errors occur the DLL is not likely to be resilient.
 * 3. Update callbacks are not implemented.
 * 4. Folder lists are calculated on the fly. This is very inefficient.
 */

#if defined( WINNT ) || defined( WIN95 )
#pragma warning ( disable : 4001 4115 4201 4209 4214 4514 )
#include <windows.h>
#pragma warning ( disable : 4001 4115 4201 4209 4214 4514 4100 4127 4152 4706)
#else
#include "windows.h"
#endif

#include <stdio.h>

#ifndef MAIN
#define MAIN
#endif

#include "efm.h"

/*
 * GMS error codes
 */
#define S_SUCCESS                   0

#define E_OUT_MEMORY                26
#define E_WRITE_FILE_FAILURE        33
#define E_READ_FILE_FAILURE         34
#define E_NO_FILE                   56
#define E_BAD_VALUE                 112
#define E_CREATE_DIR				114
#define E_NO_FOLDER                 125
#define E_INTERNAL_ERROR            231
#define E_NOT_IMPLEMENTED           241

/*
 * Filenames and locations
 */
#define EFMBaseDir				"C:\\EFM"

#define EFMSubscribedDir        "subscribed"

#define EFMNextUIDFilename		"nextuid.txt"
#define EFMUIDValidityFilename  "uidvalidity.txt"
#define EFMModifiedFilename     "modified.txt"
#define EFMSizeFilename         "size.txt"

#define EFM_MAILBOX_EXT			"mbx"
#define EFM_ATTRIBUTES_EXT		"idx"

/*
 * Timeout on opening files
 */
#define EFM_LOCK_TIMEOUT 30


/*
 * Miscellaneous
 */
#define EFM_MAX_PARAMETER   1024
#define EFM_MAX_LINE        10240

/*
 * Private Types
 */
typedef struct {
	char user[EFM_MAX_EMAIL_ADDRESS];
	char domain[EFM_MAX_EMAIL_ADDRESS];
	int serviceType;
} EFM_SESSION;

typedef struct {
	char user[EFM_MAX_EMAIL_ADDRESS];
	char domain[EFM_MAX_EMAIL_ADDRESS];
	char foldername[EFM_MAX_FILENAME];
	char indexfilename[EFM_MAX_FILENAME];
	int serviceType;
	int count;
} EFM_FOLDER;

/*
 * Entry and Exit Trace
 */
static LPFN_EFM_UPDATE_CALLBACK_FN efmUpdateCallback = NULL;

void *LogHandle = NULL;
void (*DLLLogMes) (void *Log, int Level, int id, const char *fmt, ...) = NULL;

BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
	if (reason == DLL_PROCESS_ATTACH)
	{
		return TRUE;
	}
	else if (reason == DLL_PROCESS_DETACH)
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

char *CreateUDIL(char *result, size_t size)
{
	static LONG g = 0;
	SYSTEMTIME now;

	GetSystemTime(&now);

	InterlockedIncrement(&g);

	sprintf(result, "%02d%02d%02d%03d%05d", now.wHour, now.wMinute, now.wSecond, now.wMilliseconds, g % 100000);

	return result;
}

static HANDLE RetryCreateFile(LPCTSTR filename, DWORD dwDesiredAccess, DWORD dwShareMode,
			      DWORD dwCreationDistribution, DWORD dwFlagsAndAttributes, int iRetryCount)
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	int attempt = 0;

	if (iRetryCount <= 0)
		iRetryCount = 1;
	iRetryCount *= 10;

	while (attempt < iRetryCount && hFile == INVALID_HANDLE_VALUE)
	{
		hFile = CreateFile(filename, dwDesiredAccess, dwShareMode, NULL,
				   dwCreationDistribution, dwFlagsAndAttributes, NULL);
		attempt++;
		if (iRetryCount > 10 && hFile == INVALID_HANDLE_VALUE)
		{
			DWORD error = GetLastError();

			if (error == ERROR_PATH_NOT_FOUND)
			{
				char Directory[EFM_MAX_FILENAME];
				char *End;

				strcpy(Directory, filename);
				End = strrchr(Directory, '\\');
				if (End)
				{
					*End = '\0';

					CreateDirectory(Directory, NULL);
				}
			}

			if (attempt < iRetryCount)
				Sleep(100);
		}
	}

	return (hFile);
}

static int RecurseCreateDirectory(const char *path)
{
	char Buffer[EFM_MAX_FILENAME];
	char *End;
	BOOL OK = TRUE;

	strcpy(Buffer, path);

	End = Buffer + strlen(Buffer) - 1;
	if (*End == '\\')
		*End = '\0';

	End = Buffer;
	if (*End == '\\')
		End++;
	End = strchr(End, '\\');
	if (End)
		if (*(End + 1) == '\\' || *(End - 1) == ':')
			End = strchr(End + 2, '\\');

	if (End)
		*End = '\0';

	while (End)
	{
		OK = CreateDirectory(Buffer, NULL);

		*End = '\\';
		End++;

		End = strchr(End, '\\');

		if (End)
			*End = '\0';
	}

	OK = CreateDirectory(Buffer, NULL);

	return (OK);
}


static void EFMGetMailboxFilenameFromUID(char *user, char *domain, char *foldername, unsigned int uid, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%d.%s", EFMBaseDir, domain, user, foldername, uid, EFM_MAILBOX_EXT);
}

static void EFMGetAttributesFilenameFromUID(char *user, char *domain, char *foldername, unsigned int uid, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%d.%s", EFMBaseDir, domain, user, foldername, uid, EFM_ATTRIBUTES_EXT);
}

static void EFMGetUIDValidityFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%s", EFMBaseDir, domain, user, foldername, EFMUIDValidityFilename);
}

static void EFMGetNextUIDFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%s", EFMBaseDir, domain, user, foldername, EFMNextUIDFilename);
}

static void EFMGetModifiedFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%s", EFMBaseDir, domain, user, foldername, EFMModifiedFilename);
}

static void EFMGetSizeFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%s", EFMBaseDir, domain, user, foldername, EFMSizeFilename);
}

static void EFMGetFolderFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s", EFMBaseDir, domain, user, foldername);
}

static void EFMGetFolderSearchFilename(char *user, char *domain, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\*", EFMBaseDir, domain, user);
}

static void EFMGetSubFolderSearchFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s.*", EFMBaseDir, domain, user, foldername);
}

static void EFMGetSubscribedFolderFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\%s", EFMBaseDir, domain, user, EFMSubscribedDir, foldername);
}

static void EFMGetSubscribedFolderSearchFilename(char *user, char *domain, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\*", EFMBaseDir, domain, user, EFMSubscribedDir);
}

static void EFMGetMailboxSearchFilename(char *user, char *domain, char *foldername, char *filename)
{
	sprintf(filename, "%s\\%s\\%s\\%s\\*.%s", EFMBaseDir, domain, user, foldername, EFM_MAILBOX_EXT);
}

static int EFMGetUIDValidity(int id, char *user, char *domain, char *foldername, BOOL reset, unsigned int *uidvalidity)
{
	BOOL ok;
	int status = S_SUCCESS;
	HANDLE hFile;
	char filename[EFM_MAX_FILENAME];
	char buffer[EFM_MAX_PARAMETER];
	DWORD len = sizeof(buffer);

	*uidvalidity = 0;

	EFMGetUIDValidityFilename(user, domain, foldername, filename);

	if (reset)
	{
		DeleteFile(filename);
	}

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		ok = ReadFile(hFile, buffer, sizeof(buffer), &len, FALSE);

		if (ok)
		{
			if (len != 0)
			{
				*(buffer + len) = '\0';

				*uidvalidity = atoi(buffer);
			}
			else
			{
				SYSTEMTIME st;

				GetSystemTime(&st);

				len = sprintf(buffer, "%02u%1u%02u%05u",
					      st.wMonth, st.wYear % 10, st.wDay, (st.wHour * 60 + st.wMinute) * 60 + st.wSecond);

				*uidvalidity = atoi(buffer);

				SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
				SetEndOfFile(hFile);

				ok = WriteFile(hFile, buffer, len, &len, FALSE);

				if (!ok)
				{
					status = E_WRITE_FILE_FAILURE;
				}
			}
		}
		else
		{
			status = E_READ_FILE_FAILURE;
		}

		CloseHandle(hFile);

		if (status != S_SUCCESS)
		{
			DeleteFile(filename);
		}
	}
	else
	{
		status = E_NO_FILE;
	}

	return status;
}

static int EFMGetNextUID(int id, char *user, char *domain, char *foldername, BOOL allocate, unsigned int *uid)
{
	BOOL ok;
	int status = S_SUCCESS;
	HANDLE hFile;
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	char filename[EFM_MAX_FILENAME];
	char buffer[EFM_MAX_PARAMETER];
	DWORD len = sizeof(buffer);
	unsigned int fileuid = 0;
	unsigned int maxuid = 0;
	BOOL save = allocate;
	unsigned int uidvalidity;

	*uid = 0;

	EFMGetNextUIDFilename(user, domain, foldername, filename);

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		ok = ReadFile(hFile, buffer, sizeof(buffer), &len, FALSE);

		if (ok)
		{
			if (len != 0)
			{
				*(buffer + len) = '\0';

				*uid = atoi(buffer);
			}
			else
			{
				// Rebuild the max_uid from the files
				EFMGetMailboxSearchFilename(user, domain, foldername, filename);

				hFindFile = FindFirstFile(filename, &finddata);

				if (hFindFile != INVALID_HANDLE_VALUE)
				{
					ok = TRUE;

					while (ok)
					{
						fileuid = atoi(finddata.cFileName);

						if (fileuid > maxuid)
						{
							maxuid = fileuid;
						}

						ok = FindNextFile(hFindFile, &finddata);
					}

					FindClose(hFindFile);

					*uid = maxuid + 1;
				}
				else
				{
					*uid = 1;
				}

				// Need to reset UIDValidity
				status = EFMGetUIDValidity(id, user, domain, foldername, TRUE, &uidvalidity);

				save = TRUE;
			}
		}
		else
		{
			status = E_READ_FILE_FAILURE;
		}

		if (status == S_SUCCESS)
		{
			maxuid = *uid;

			if (allocate)
			{
				maxuid++;
			}

			if (save)
			{
				len = sprintf(buffer, "%d", maxuid);

				SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
				SetEndOfFile(hFile);

				ok = WriteFile(hFile, buffer, len, &len, FALSE);

				if (!ok)
				{
					status = E_WRITE_FILE_FAILURE;
				}
			}
		}

		CloseHandle(hFile);

		if (status != S_SUCCESS)
		{
			DeleteFile(filename);
		}
	}
	else
	{
		status = E_NO_FILE;
	}

	return status;
}

static int EFMCleanupFolderFiles(char *user, char *domain, char *foldername)
{
	int status = S_SUCCESS;
	char filename[EFM_MAX_FILENAME];

	EFMGetUIDValidityFilename(user, domain, foldername, filename);
	DeleteFile(filename);

	EFMGetNextUIDFilename(user, domain, foldername, filename);
	DeleteFile(filename);

	EFMGetModifiedFilename(user, domain, foldername, filename);
	DeleteFile(filename);

	EFMGetSizeFilename(user, domain, foldername, filename);
	DeleteFile(filename);

	EFMGetFolderFilename(user, domain, foldername, filename);
	RemoveDirectory(filename);

	return status;
}

static int EFMCleanupUIDFiles(char *user, char *domain, char *foldername, unsigned int uid)
{
	int status = S_SUCCESS;
	char filename[EFM_MAX_FILENAME];

	EFMGetMailboxFilenameFromUID(user, domain, foldername, uid, filename);
	DeleteFile(filename);

	EFMGetAttributesFilenameFromUID(user, domain, foldername, uid, filename);
	DeleteFile(filename);

	return status;
}

static int EFMWriteAttributes(char *user, char *domain, char *foldername, int uid, char *udil, char *flags)
{
	BOOL ok;
	int status = S_SUCCESS;
	char filename[EFM_MAX_FILENAME];
	HANDLE hFile;
	char buffer[EFM_MAX_LINE];
	DWORD len;

	EFMGetAttributesFilenameFromUID(user, domain, foldername, uid, filename);

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		len = sprintf(buffer, "%s %s", udil, flags);

		SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
		SetEndOfFile(hFile);

		ok = WriteFile(hFile, buffer, len, &len, FALSE);

		if (!ok)
		{
			status = E_WRITE_FILE_FAILURE;
		}

		CloseHandle(hFile);
	}
	else
	{
		status = E_NO_FILE;
	}

	return status;
}

static int EFMGetMessageAttributes(char *user,
				   char *domain,
				   char *foldername,
				   int uid, char *udil, int udilSize, char *flags, int flagsSize, unsigned long *size)
{
	BOOL ok;
	int status = S_SUCCESS;
	char mailboxFilename[EFM_MAX_FILENAME];
	char attributesFilename[EFM_MAX_FILENAME];
	HANDLE hFile;
	char buffer[EFM_MAX_LINE];
	DWORD len = sizeof(buffer);
	char *ptr;

	EFMGetMailboxFilenameFromUID(user, domain, foldername, uid, mailboxFilename);
	EFMGetAttributesFilenameFromUID(user, domain, foldername, uid, attributesFilename);

	if (size != NULL)
	{
		hFile = RetryCreateFile(mailboxFilename, GENERIC_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			*size = GetFileSize(hFile, NULL);

			CloseHandle(hFile);
		}
		else
		{
			status = E_NO_FILE;
		}
	}

	if (udil != NULL || flags != NULL)
	{
		if (status == S_SUCCESS)
		{
			hFile =
			    RetryCreateFile(attributesFilename, GENERIC_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
					    EFM_LOCK_TIMEOUT);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				ok = ReadFile(hFile, buffer, sizeof(buffer), &len, FALSE);

				if (ok)
				{
					*(buffer + len) = '\0';

					ptr = strchr(buffer, ' ');

					if (ptr != NULL)
					{
						*ptr = '\0';
						ptr++;

						if (udil != NULL)
						{
							strcpy(udil, buffer);
						}

						if (flags != NULL)
						{
							strcpy(flags, ptr);
						}
					}
				}
				else
				{
					status = E_READ_FILE_FAILURE;
				}

				CloseHandle(hFile);
			}
		}
	}

	return status;
}

static int EFMUpdateModifiedTime(char *user, char *domain, char *foldername, FILETIME * lastWriteTime)
{
	BOOL ok;
	int status = S_SUCCESS;
	HANDLE hFile;
	char filename[EFM_MAX_FILENAME];
	SYSTEMTIME st;
	FILETIME ft;

	if (lastWriteTime == NULL)
	{
		GetSystemTime(&st);
		SystemTimeToFileTime(&st, &ft);

		lastWriteTime = &ft;
	}

	EFMGetModifiedFilename(user, domain, foldername, filename);

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		ok = SetFileTime(hFile, NULL, NULL, &ft);

		if (!ok)
		{
			status = E_WRITE_FILE_FAILURE;
		}

		CloseHandle(hFile);
	}

	return status;
}

static int EFMUpdateSize(char *user, char *domain, char *foldername, int size)
{
	BOOL ok;
	int status = S_SUCCESS;
	HANDLE hFile;
	char filename[EFM_MAX_FILENAME];
	char buffer[EFM_MAX_PARAMETER];
	DWORD len = sizeof(buffer);

	EFMGetSizeFilename(user, domain, foldername, filename);

	hFile = RetryCreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, EFM_LOCK_TIMEOUT);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		len = sprintf(buffer, "%d", size);

		ok = WriteFile(hFile, buffer, len, &len, FALSE);

		if (!ok)
		{
			status = E_WRITE_FILE_FAILURE;
		}

		CloseHandle(hFile);
	}

	return status;
}

static void EFMDeleteSize(char *user, char *domain, char *foldername)
{
	char filename[EFM_MAX_FILENAME];

	EFMGetSizeFilename(user, domain, foldername, filename);

	DeleteFile(filename);
}

static BOOL EFMIsFlag(char *status, const char *y)
{
	int len = strlen(y);
	char *start = status;
	char *end = strchr(y, ',');
	BOOL found = FALSE;

	while (start && *start && !found)
	{
		char *prev = start - 1;

		found = (start == status || *prev == ',') && (end || *(start + len) == ',') && (strnicmp(start, y, len) == 0);
		if (!found)
			start = strstr(start + 1, y);
	}

	return (found);
}

static BOOL EFMIsFlags(char *status, const char *y)
{
	BOOL found = FALSE;
	char *tok;
	char searchStatus[EFM_MAX_LINE];

	strcpy(searchStatus, y);

	tok = strtok(searchStatus, ",");

	while (tok != NULL && !found)
	{
		found = EFMIsFlag(status, tok);

		tok = strtok(NULL, ",");
	}

	return (found);
}

static void EFMFreeFolderList(EFM_FOLDER_LIST ** list)
{
	EFM_FOLDER_LIST *current;
	EFM_FOLDER_LIST *next;

	current = *list;

	while (current != NULL)
	{
		next = current->next;

		if (current->foldername != NULL)
		{
			free(current->foldername);
		}

		free(current);

		current = next;
	}

	*list = NULL;
}

static int EFMAppendFolder(int id,
			   char *user, char *domain,
			   char *foldername, char *filename, char *flags, unsigned int *uid, char *udil, int size)
{
	BOOL ok;
	int status = S_SUCCESS;
	unsigned int newuid = 0;
	char newudil[EFM_MAX_UDIL];
	char mailboxFilename[EFM_MAX_FILENAME];

	status = EFMGetNextUID(id, user, domain, foldername, TRUE, &newuid);

	if (status == S_SUCCESS)
	{
		EFMGetMailboxFilenameFromUID(user, domain, foldername, newuid, mailboxFilename);

		ok = CopyFile(filename, mailboxFilename, TRUE);

		if (!ok)
		{
			status = E_WRITE_FILE_FAILURE;
		}
	}

	if (status == S_SUCCESS)
	{
		CreateUDIL(newudil, sizeof(newudil));

		status = EFMWriteAttributes(user, domain, foldername, newuid, newudil, flags);
	}

	if (status == S_SUCCESS)
	{
		if (uid)
		{
			*uid = newuid;
		}

		if (udil)
		{
			strcpy(udil, newudil);
		}

		EFMDeleteSize(user, domain, foldername);

		EFMUpdateModifiedTime(user, domain, foldername, NULL);
	}
	else
	{
		EFMCleanupUIDFiles(user, domain, foldername, newuid);
	}

	return status;
}

int EFMConfigure(int id, void *logMes, void *logHandle)
{
	int status = E_BAD_VALUE;

	if (logMes != NULL)
	{
		DLLLogMes = logMes;
		LogHandle = logHandle;

		status = S_SUCCESS;
	}

	return (status);
}

int EFMClose(int id)
{
	int status = S_SUCCESS;

	// No action

	return status;
}

int EFMCreateSession(int id, char *user, char *domain, int serviceType, EFM_SESSION_HANDLE * sessionHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	*sessionHandle = EFM_INVALID_SESSION_HANDLE;

	session = malloc(sizeof(EFM_SESSION));

	if (session != NULL)
	{
		strcpy(session->user, user);
		strcpy(session->domain, domain);
		session->serviceType = serviceType;

		*sessionHandle = (EFM_SESSION_HANDLE) session;
	}
	else
	{
		status = E_OUT_MEMORY;
	}

	return status;
}

int EFMCloseSession(int id, EFM_SESSION_HANDLE * sessionHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (*sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) * sessionHandle;

		free((void *)*sessionHandle);
		*sessionHandle = EFM_INVALID_SESSION_HANDLE;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMKeepAlive(int id, EFM_SESSION_HANDLE sessionHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		// No action
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMExpungedMessageAccessAllowed(int id, EFM_SESSION_HANDLE sessionHandle, BOOL *allowed)
{
	int status = S_SUCCESS;

	*allowed = TRUE;

	return status;
}

/*
 * Messages
 */
int EFMGetMessage(int id, EFM_FOLDER_HANDLE folderHandle, unsigned int uid, char *filename, size_t size)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		EFMGetMailboxFilenameFromUID(folder->user, folder->domain, folder->foldername, uid, filename);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseMessage(int id, EFM_FOLDER_HANDLE folderHandle, unsigned int uid, char *filename)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		*filename = '\0';
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCopyMessages(int id, EFM_FOLDER_HANDLE folderHandle, char *foldername, EFM_MESSAGE_LIST * messageList)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;
	EFM_MESSAGE_LIST *current;
	unsigned int uid;
	char udil[EFM_MAX_UDIL];
	char flags[EFM_MAX_FLAGS];
	char filename[EFM_MAX_FILENAME];

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		current = messageList;

		while (current != NULL && status == S_SUCCESS)
		{
			status = EFMGetMessageAttributes(folder->user, folder->domain, folder->foldername,
							 current->uid, NULL, 0, flags, sizeof(flags), NULL);

			if (status == S_SUCCESS)
			{
				EFMGetMailboxFilenameFromUID(folder->user, folder->domain, folder->foldername, current->uid,
							     filename);

				status = EFMAppendFolder(id,
							 folder->user, folder->domain,
							 foldername, filename, flags, &uid, udil, sizeof(udil));

				// Update UIDs
				current->uid = uid;
				strcpy(current->udil, udil);
			}

			current = current->next;
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMSetMessageFlags(int id, EFM_FOLDER_HANDLE folderHandle, unsigned int uid, char *flags)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;
	char udil[EFM_MAX_UDIL];

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// We should cache these until a commit but for now we simplify and write directly.
		status = EFMGetMessageAttributes(folder->user, folder->domain, folder->foldername,
						 uid, udil, sizeof(udil), NULL, 0, NULL);

		if (status == S_SUCCESS)
		{
			status = EFMWriteAttributes(folder->user, folder->domain, folder->foldername, uid, udil, flags);

		}

		if (status == S_SUCCESS)
		{
			EFMUpdateModifiedTime(folder->user, folder->domain, folder->foldername, NULL);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

/*
 * Folders
 */
int EFMIsExternalFolder(int id, char *user, char *domain, int serviceType, BOOL *external)
{
	int status = S_SUCCESS;

	*external = TRUE;

	return status;
}

int EFMGetFolderUIDValidity(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, unsigned int *uidvalidity)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMGetUIDValidity(id, session->user, session->domain, foldername, FALSE, uidvalidity);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderNextUID(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, unsigned int *uid)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMGetNextUID(id, session->user, session->domain, foldername, FALSE, uid);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}


int EFMExistFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, BOOL *exists)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	WIN32_FIND_DATA finddata;
	HANDLE hFindFile;
	char filename[EFM_MAX_FILENAME];

	*exists = FALSE;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetFolderFilename(session->user, session->domain, foldername, filename);

		hFindFile = FindFirstFile(filename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
			{
				*exists = TRUE;
			}

			FindClose(hFindFile);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCreateFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	BOOL exists = FALSE;
	char directory[EFM_MAX_FILENAME];
	unsigned int uidvalidity;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMExistFolder(id, sessionHandle, foldername, &exists);

		if (status == S_SUCCESS && !exists)
		{
			EFMGetFolderFilename(session->user, session->domain, foldername, directory);

			if (RecurseCreateDirectory(directory))
			{
				// Set the UID Validity up
				status = EFMGetUIDValidity(id, session->user, session->domain, foldername, FALSE, &uidvalidity);
			}
			else
			{
				status = E_CREATE_DIR;
			}
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMDeleteFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char searchFilename[EFM_MAX_FILENAME];
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	unsigned int uid;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetMailboxSearchFilename(session->user, session->domain, foldername, searchFilename);

		hFindFile = FindFirstFile(searchFilename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			ok = TRUE;

			while (ok)
			{
				if (strcmp(finddata.cFileName, ".") == 0 || strcmp(finddata.cFileName, "..") == 0)
				{
					// Ignore
				}
				else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
				{
					uid = atoi(finddata.cFileName);

					EFMCleanupUIDFiles(session->user, session->domain, foldername, uid);
				}

				ok = FindNextFile(hFindFile, &finddata);
			}

			FindClose(hFindFile);
		}

		EFMCleanupFolderFiles(session->user, session->domain, foldername);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char filename[EFM_MAX_FILENAME];
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	EFM_FOLDER_LIST *pNewFolder;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetFolderSearchFilename(session->user, session->domain, filename);

		hFindFile = FindFirstFile(filename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			ok = TRUE;

			while (ok)
			{
				if (strcmp(finddata.cFileName, ".") == 0 || strcmp(finddata.cFileName, "..") == 0)
				{
					// Ignore
				}
				else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
				{
					pNewFolder = malloc(sizeof(EFM_FOLDER_LIST));

					if (pNewFolder == NULL)
					{
						status = E_OUT_MEMORY;
					}
					else
					{
						memset(pNewFolder, 0, sizeof(EFM_FOLDER_LIST));

						pNewFolder->attributes = 0;
						pNewFolder->foldername = strdup(finddata.cFileName);

						if (pNewFolder->foldername == NULL)
						{
							status = E_OUT_MEMORY;
						}
						else
						{
							pNewFolder->next = *list;
							*list = pNewFolder;
						}

						if (status != S_SUCCESS)
						{
							free(pNewFolder);
						}
					}
				}

				ok = FindNextFile(hFindFile, &finddata);
			}

			FindClose(hFindFile);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMFreeFolderList(list);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderModified(int id, EFM_SESSION_HANDLE sessionHandle, char *folderName, char *modifiedStr, int size)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	BOOL ok;
	HANDLE hFile;
	char modifiedFilename[EFM_MAX_FILENAME];
	char searchFilename[EFM_MAX_FILENAME];
	FILETIME lastWriteTime = { 0 };
	SYSTEMTIME st;
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	BOOL found;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetModifiedFilename(session->user, session->domain, folderName, modifiedFilename);

		hFile = RetryCreateFile(modifiedFilename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			ok = GetFileTime(hFile, NULL, NULL, &lastWriteTime);

			if (!ok)
			{
				status = E_WRITE_FILE_FAILURE;
			}

			CloseHandle(hFile);
		}
		else
		{
			found = FALSE;

			EFMGetMailboxSearchFilename(session->user, session->domain, folderName, searchFilename);

			hFindFile = FindFirstFile(searchFilename, &finddata);

			if (hFindFile != INVALID_HANDLE_VALUE)
			{
				ok = TRUE;

				while (ok)
				{
					if (strcmp(finddata.cFileName, ".") == 0)
					{
						if (CompareFileTime(&finddata.ftLastWriteTime, &lastWriteTime) > 0)
						{
							lastWriteTime = finddata.ftLastWriteTime;
							found = TRUE;
						}
					}
					else if (strcmp(finddata.cFileName, "..") == 0)
					{
						// Ignore
					}
					else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
					{
						if (CompareFileTime(&finddata.ftLastWriteTime, &lastWriteTime) > 0)
						{
							lastWriteTime = finddata.ftLastWriteTime;
							found = TRUE;
						}
					}

					ok = FindNextFile(hFindFile, &finddata);
				}

				FindClose(hFindFile);
			}

			if (found)
			{
				/*
				 * Cache the modified time
				 */
				EFMUpdateModifiedTime(session->user, session->domain, folderName, &lastWriteTime);
			}
		}

		if (status == S_SUCCESS)
		{
			if (FileTimeToSystemTime(&lastWriteTime, &st))
			{
				sprintf(modifiedStr,
					"%04d-%02d-%02d %02d:%02d:%02d",
					st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond);
			}
			else
			{
				status = E_INTERNAL_ERROR;
			}
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderSize(int id, EFM_SESSION_HANDLE sessionHandle, char *folderName, DWORD * size)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	BOOL ok;
	HANDLE hFile;
	char sizeFilename[EFM_MAX_FILENAME];
	char searchFilename[EFM_MAX_FILENAME];
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	char buffer[EFM_MAX_FILENAME];
	DWORD len = sizeof(buffer);
	unsigned long currentSize = 0;
	unsigned int uid;
	char flags[EFM_MAX_FLAGS];

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetSizeFilename(session->user, session->domain, folderName, sizeFilename);

		hFile = RetryCreateFile(sizeFilename, GENERIC_READ | GENERIC_WRITE, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			ok = ReadFile(hFile, buffer, sizeof(buffer), &len, FALSE);

			if (ok)
			{
				*(buffer + len) = '\0';

				*size = atoi(buffer);
			}
			else
			{
				status = E_READ_FILE_FAILURE;
			}

			CloseHandle(hFile);
		}
		else
		{
			EFMGetMailboxSearchFilename(session->user, session->domain, folderName, searchFilename);

			hFindFile = FindFirstFile(searchFilename, &finddata);

			if (hFindFile != INVALID_HANDLE_VALUE)
			{
				ok = TRUE;

				while (ok)
				{
					if (strcmp(finddata.cFileName, ".") == 0 || strcmp(finddata.cFileName, "..") == 0)
					{
						// Ignore
					}
					else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
					{
						uid = atoi(finddata.cFileName);

						status = EFMGetMessageAttributes(session->user, session->domain, folderName,
										 uid, NULL, 0, flags, sizeof(flags), NULL);

						if (status == S_SUCCESS)
						{
							if (!EFMIsFlag(flags, "EXPUNGED"))
							{
								currentSize += finddata.nFileSizeLow;
							}
						}
					}

					ok = FindNextFile(hFindFile, &finddata);
				}

				FindClose(hFindFile);
			}

			/*
			 * Cache the size
			 */
			if (status == S_SUCCESS)
			{
				EFMUpdateSize(session->user, session->domain, folderName, currentSize);

				*size = currentSize;
			}
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMFolderCountMessagesOfType(int id,
				 EFM_SESSION_HANDLE sessionHandle,
				 char *foldername, char *includeFlags, char *excludeFlags, int *messageCount)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char searchFilename[EFM_MAX_FILENAME];
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	unsigned int uid;
	char flags[EFM_MAX_FLAGS];

	*messageCount = 0;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetMailboxSearchFilename(session->user, session->domain, foldername, searchFilename);

		hFindFile = FindFirstFile(searchFilename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			ok = TRUE;

			while (ok)
			{
				if (strcmp(finddata.cFileName, ".") == 0 || strcmp(finddata.cFileName, "..") == 0)
				{
					// Ignore
				}
				else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
				{
					uid = atoi(finddata.cFileName);

					status = EFMGetMessageAttributes(session->user, session->domain, foldername,
									 uid, NULL, 0, flags, sizeof(flags), NULL);

					if (status == S_SUCCESS)
					{
						if (excludeFlags == NULL ||
						    *excludeFlags == '\0' || !EFMIsFlags(flags, excludeFlags))
						{
							if (includeFlags == NULL ||
							    *includeFlags == '\0' || EFMIsFlags(flags, includeFlags))
							{
								(*messageCount)++;
							}
						}
					}
				}

				ok = FindNextFile(hFindFile, &finddata);
			}

			FindClose(hFindFile);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMRenameFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, char *newfoldername, BOOL renameSubFolders)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char filename[EFM_MAX_FILENAME];
	char newFilename[EFM_MAX_FILENAME];
	char inferiorFilename[EFM_MAX_FILENAME];
	DWORD dwAttributes;
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	unsigned int len;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetFolderFilename(session->user, session->domain, foldername, filename);
		EFMGetFolderFilename(session->user, session->domain, newfoldername, newFilename);

		if ((dwAttributes = GetFileAttributes(filename)) != -1)
		{
			if ((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
			{
				ok = MoveFile(filename, newFilename);

				if (ok)
				{
					EFMUpdateModifiedTime(session->user, session->domain, newfoldername, NULL);
				}
			}
		}

		if (renameSubFolders)
		{
			EFMGetSubFolderSearchFilename(session->user, session->domain, foldername, filename);

			hFindFile = FindFirstFile(filename, &finddata);

			if (hFindFile != INVALID_HANDLE_VALUE)
			{
				len = strlen(foldername);

				ok = TRUE;

				while (ok)
				{
					if (strcmp(finddata.cFileName, ".") == 0 || strcmp(finddata.cFileName, "..") == 0)
					{
						// Ignore
					}
					else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
					{
						if (strlen(finddata.cFileName) >= (len + 2))
						{
							strcpy(inferiorFilename, newfoldername);
							strcat(inferiorFilename, ".");
							strcat(inferiorFilename, finddata.cFileName + len + 1);

							EFMGetFolderFilename(session->user, session->domain, finddata.cFileName,
									     filename);
							EFMGetFolderFilename(session->user, session->domain, inferiorFilename,
									     newFilename);

							ok = MoveFile(filename, newFilename);

							if (ok)
							{
								EFMUpdateModifiedTime(session->user,
										      session->domain,
										      finddata.cFileName + len + 1, NULL);
							}
						}
					}

					ok = FindNextFile(hFindFile, &finddata);
				}

				FindClose(hFindFile);
			}
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMOpenFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, int flags, EFM_FOLDER_HANDLE * folderHandle)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	EFM_FOLDER *folder;

	if (flags == EFM_FOLDER_OPEN_FULL)
	{
		*folderHandle = EFM_INVALID_FOLDER_HANDLE;

		if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
		{
			session = (EFM_SESSION *) sessionHandle;

			folder = malloc(sizeof(EFM_FOLDER));

			if (folder != NULL)
			{
				memset(folder, 0, sizeof(EFM_FOLDER));

				strcpy(folder->user, session->user);
				strcpy(folder->domain, session->domain);
				strcpy(folder->foldername, foldername);
				folder->serviceType = session->serviceType;
				folder->count = 1;

				*folderHandle = (EFM_FOLDER_HANDLE) folder;
			}
			else
			{
				status = E_OUT_MEMORY;
			}
		}
		else
		{
			status = E_INTERNAL_ERROR;
		}
	}
	else if (flags == EFM_FOLDER_OPEN_INFORMATION)
	{
		if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
		{
			session = (EFM_SESSION *) sessionHandle;

			if (*folderHandle != EFM_INVALID_FOLDER_HANDLE)
			{
				folder = (EFM_FOLDER *) * folderHandle;

				folder->count++;
			}
			else
			{
				status = E_INTERNAL_ERROR;
			}
		}
		else
		{
			status = E_INTERNAL_ERROR;
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetFolderIndex(int id, EFM_FOLDER_HANDLE folderHandle, EFM_MESSAGE_LIST ** messageList)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_FOLDER *folder;
	char filename[EFM_MAX_FILENAME];
	unsigned int uid;
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	EFM_MESSAGE_LIST *currentMessageList = NULL;
	EFM_MESSAGE_LIST *lastMessageList = NULL;
	unsigned int number = 1;

	*messageList = NULL;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		EFMGetMailboxSearchFilename(folder->user, folder->domain, folder->foldername, filename);

		// Assume UIDs in order
		hFindFile = FindFirstFile(filename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			ok = TRUE;

			while (ok)
			{
				uid = atoi(finddata.cFileName);

				currentMessageList = malloc(sizeof(EFM_MESSAGE_LIST));

				if (currentMessageList != NULL)
				{
					currentMessageList->number = number++;
					currentMessageList->uid = uid;

					status = EFMGetMessageAttributes(folder->user, folder->domain, folder->foldername,
									 uid,
									 currentMessageList->udil, sizeof(currentMessageList->udil),
									 currentMessageList->flags,
									 sizeof(currentMessageList->flags),
									 &currentMessageList->size);

					currentMessageList->next = NULL;

					if (lastMessageList != NULL)
					{
						lastMessageList->next = currentMessageList;
					}
					else
					{
						*messageList = currentMessageList;
					}

					lastMessageList = currentMessageList;
				}
				else
				{
					status = E_OUT_MEMORY;
				}

				ok = FindNextFile(hFindFile, &finddata);
			}

			FindClose(hFindFile);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseFolderIndex(int id, EFM_FOLDER_HANDLE folderHandle, EFM_MESSAGE_LIST ** messageList)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;
	EFM_MESSAGE_LIST *currentMessageList = NULL;
	EFM_MESSAGE_LIST *nextMessageList = NULL;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		currentMessageList = *messageList;

		while (currentMessageList != NULL)
		{
			nextMessageList = currentMessageList->next;

			free(currentMessageList);

			currentMessageList = nextMessageList;
		}

		*messageList = NULL;
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseFolder(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_HANDLE * folderHandle, int flags)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	EFM_FOLDER *folder;


	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;
	}

	if (*folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) * folderHandle;

		if (flags == EFM_FOLDER_CLOSE_FULL)
		{
			free((void *)*folderHandle);
			*folderHandle = EFM_INVALID_FOLDER_HANDLE;
		}
		else if (flags == EFM_FOLDER_CLOSE_INFORMATION)
		{
			folder->count--;
		}
		else
		{
			status = E_INTERNAL_ERROR;
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCommitFolder(int id, EFM_FOLDER_HANDLE folderHandle)
{
	int status = S_SUCCESS;
	EFM_FOLDER *folder;

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		// Commit is not supported, status updates written immediately.
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMPurgeFolder(int id, EFM_FOLDER_HANDLE folderHandle)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_FOLDER *folder;
	char filename[EFM_MAX_FILENAME];
	HANDLE hFindFile;
	WIN32_FIND_DATA finddata;
	unsigned int uid;
	char flags[EFM_MAX_FLAGS];

	if (folderHandle != EFM_INVALID_FOLDER_HANDLE)
	{
		folder = (EFM_FOLDER *) folderHandle;

		EFMGetMailboxSearchFilename(folder->user, folder->domain, folder->foldername, filename);

		hFindFile = FindFirstFile(filename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			ok = TRUE;

			while (ok)
			{
				uid = atoi(finddata.cFileName);

				status = EFMGetMessageAttributes(folder->user, folder->domain, folder->foldername,
								 uid, NULL, 0, flags, sizeof(flags), NULL);

				if (EFMIsFlag(flags, "DELETED"))
				{
					EFMCleanupUIDFiles(folder->user, folder->domain, folder->foldername, uid);
				}

				ok = FindNextFile(hFindFile, &finddata);
			}

			FindClose(hFindFile);

			EFMDeleteSize(folder->user, folder->domain, folder->foldername);

			EFMUpdateModifiedTime(folder->user, folder->domain, folder->foldername, NULL);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMStoreFolder(int id,
		   EFM_SESSION_HANDLE sessionHandle,
		   char *foldername, char *filename, char *flags, unsigned int *uid, char *udil, int size)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		status = EFMCreateFolder(id, sessionHandle, foldername);

		if (status == S_SUCCESS)
		{
			status = EFMAppendFolder(id, session->user, session->domain, foldername, filename, flags, uid, udil, size);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

/*
 * Folder Locking
 */
int EFMClientFolderLock(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, EFM_LOCK_HANDLE *lockHandle)
{
	int status = S_SUCCESS;

	// No locking
	*lockHandle = EFM_INVALID_LOCK_HANDLE;

	return status;
}

int EFMServiceFolderLock(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, EFM_LOCK_HANDLE *lockHandle)
{
	int status = S_SUCCESS;

	// No locking
	*lockHandle = EFM_INVALID_LOCK_HANDLE;

	return status;
}

int EFMCloseFolderLock(int id, EFM_LOCK_HANDLE *lockHandle)
{
	int status = S_SUCCESS;

	// No locking
	*lockHandle = EFM_INVALID_LOCK_HANDLE;

	return status;
}

int EFMGMSClientFolderLockRequired(int id, BOOL *required)
{
	int status = S_SUCCESS;

	// We do no locking so we should require GMS to do the locking
	*required = TRUE;

	return status;
}

int EFMGMSServiceFolderLockRequired(int id, BOOL *required)
{
	int status = S_SUCCESS;

	// We do no locking so we should require GMS to do the locking
	*required = TRUE;

	return status;
}

int EFMDoesClientFolderLockExist(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, BOOL *exists)
{
	int status = S_SUCCESS;

	// No locking
	*exists = FALSE;

	return status;
}

int EFMDoesServiceFolderLockExist(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername, BOOL *exists)
{
	int status = S_SUCCESS;

	// No locking
	*exists = FALSE;

	return status;
}

/*
 * Folder Subscription
 */
int EFMSubscribeFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char filename[EFM_MAX_FILENAME];

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetSubscribedFolderFilename(session->user, session->domain, foldername, filename);

		RecurseCreateDirectory(filename);
	}

	return status;
}

int EFMUnsubscribeFolder(int id, EFM_SESSION_HANDLE sessionHandle, char *foldername)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char filename[EFM_MAX_FILENAME];

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetSubscribedFolderFilename(session->user, session->domain, foldername, filename);

		RemoveDirectory(filename);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMGetSubscribedFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	BOOL ok;
	int status = S_SUCCESS;
	EFM_SESSION *session;
	char filename[EFM_MAX_FILENAME];
	WIN32_FIND_DATA finddata;
	HANDLE hFindFile;
	EFM_FOLDER_LIST *pNewFolder;
	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMGetSubscribedFolderSearchFilename(session->user, session->domain, filename);

		hFindFile = FindFirstFile(filename, &finddata);

		if (hFindFile != INVALID_HANDLE_VALUE)
		{
			ok = TRUE;

			while (ok)
			{
				if (strcmp(finddata.cFileName, ".") == 0 || strcmp(finddata.cFileName, "..") == 0)
				{
					// Ignore
				}
				else if ((finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
				{
					pNewFolder = malloc(sizeof(EFM_FOLDER_LIST));

					if (pNewFolder == NULL)
					{
						status = E_OUT_MEMORY;
					}
					else
					{
						memset(pNewFolder, 0, sizeof(EFM_FOLDER_LIST));

						pNewFolder->attributes = 0;
						pNewFolder->foldername = strdup(finddata.cFileName);

						if (pNewFolder->foldername == NULL)
						{
							status = E_OUT_MEMORY;
						}
						else
						{
							pNewFolder->next = *list;
							*list = pNewFolder;
						}

						if (status != S_SUCCESS)
						{
							free(pNewFolder);
						}
					}
				}

				ok = FindNextFile(hFindFile, &finddata);
			}

			FindClose(hFindFile);
		}
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

int EFMCloseSubscribedFolderList(int id, EFM_SESSION_HANDLE sessionHandle, EFM_FOLDER_LIST ** list)
{
	int status = S_SUCCESS;
	EFM_SESSION *session;

	if (sessionHandle != EFM_INVALID_SESSION_HANDLE)
	{
		session = (EFM_SESSION *) sessionHandle;

		EFMFreeFolderList(list);
	}
	else
	{
		status = E_INTERNAL_ERROR;
	}

	return status;
}

/*
 * Update callback
 */
int EFMRegisterUpdateCallback(int id, LPFN_EFM_UPDATE_CALLBACK_FN updateCallback)
{
	int status = S_SUCCESS;

	status = S_SUCCESS;

	return status;
}
